"""TMTCrunch defaults and settings functions."""

__all__ = [
    "load_config",
    "load_default_config",
    "load_phospho_config",
    "format_settings",
]


from importlib.resources import files

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


def load_config(fpath: str) -> dict:
    """
    Load configuration from a file.

    :param fpath: Path to config file.
    :return: settings.
    """
    with open(fpath, "rb") as f:
        settings = tomllib.load(f)
    return settings


def load_default_config() -> dict:
    """
    Load default configuration.

    :return: settings.
    """
    config_file = files("tmtcrunch.conf").joinpath("default.toml")
    return load_config(config_file)


def load_phospho_config() -> dict:
    """
    Load default configuration for phospho-proteomics.

    :return: settings.
    """
    config_file = files("tmtcrunch.conf").joinpath("phospho.toml")
    return load_config(config_file)


def format_settings(settings: dict, pretty=True) -> str:
    """
    Return formatted representation of `settings`.

    :param settings: TMTCrunch settings.
    :param pretty: If True, add header and footer.
    :return: Formatted string.
    """
    header = "====  settings  ===="
    footer = "=" * len(header)

    settings_str = ""
    if pretty:
        settings_str += header + "\n"
    for key, value in settings.items():
        if key == "psm_group":
            settings_str += f"{key}: " + "{\n"
            for group, group_cfg in value.items():
                settings_str += f"  {group}: {group_cfg},\n"
            settings_str += "}\n"
        else:
            settings_str += f"{key}: {value}\n"
    if pretty:
        settings_str += footer
    else:
        settings_str = settings_str[:-1]
    return settings_str



"""Classes and functions to group PSMs by the prefix in FASTA headers."""

__all__ = [
    "GeneralPsmGroup",
    "PsmGroup",
    "PrimeGroupsCollection",
    "expand_collection",
    "collection_from_elements",
    "generate_prefix_collection",
    "is_derived_superset",
]


import numpy as np
import pandas as pd


def expand_collection(new_element: str, collection: list[list[str]]) -> list[list[str]]:
    """
    Create new collection of all combinations using new element and existing collection.

    :param new_element: New element.
    :param collection: Collection.
    :return: New collection of all combinations.
    """
    expanded_collection = [[new_element]] + collection
    for combination in collection:
        expanded_collection.append([new_element] + combination)
    return expanded_collection


def collection_from_elements(list_of_elements: list) -> list[list[str]]:
    """
    Create collection of all combinations of elements.

    :param list_of_elements: List of unique elements.
    :return: Collection of all combinations of elements.
    """
    collection = []
    for element in list_of_elements:
        collection = expand_collection(element, collection)
    return collection


def generate_prefix_collection(
    target_prefixes: list[str], decoy_prefix: str = "DECOY_"
) -> list[list[str]]:
    """
    Generate list of all possible prefixes for PSMs.

    :param target_prefixes: List of target prefixes.
    :param decoy_prefix: Decoy prefix, defaults to "DECOY_".
    :return: Collection of all possible prefixes.
    """
    decoy_prefixes = [decoy_prefix + tp for tp in target_prefixes]
    all_prefixes = target_prefixes + decoy_prefixes
    return collection_from_elements(all_prefixes)


def is_derived_superset(super_set: list[str], base_set: list[str]) -> bool:
    """
    Check whether list of strings is derived from the `base_set`.

    List of strings is derived from base set if all its strings starts with any and only
    with any string from the base set.

    :param super_set: List of strings.
    :param base_set: List of base strings.
    :return: True, if `super_set` is derived from the `base_set`.
    """
    mat = [[x.startswith(s) for x in super_set] for s in base_set]
    return all(np.any(mat, axis=0)) and all(np.any(mat, axis=1))


class PrimeGroupsCollection:
    """Collection of prime PSM groups."""

    def __init__(
        self,
        prefix_collection: list[list[str]],
        df_psm: pd.DataFrame = pd.DataFrame({"protein": ["none"]}),
        decoy_prefix: str = "DECOY_",
    ):
        """
        Create collection of prime PSM groups.

        :param prefix_collection: Collection of prefixes for prime groups.
        :param df_psm: Pandas DataFrame with PSMs.
        :param decoy_prefix: Decoy prefix.
        """
        self.decoy_prefix = decoy_prefix
        self.prime_groups = {
            i: {"prefixes": v} for i, v in enumerate(prefix_collection)
        }
        self.df_psm = df_psm
        for group in self.prime_groups.values():
            group["indicator"] = df_psm.protein.apply(
                lambda x: is_derived_superset(x, group["prefixes"])
            )
            group["count"] = sum(group["indicator"])
        self.total_psm = sum(group["count"] for group in self.prime_groups.values())
        self.unaccounted_psm = self.df_psm.shape[0] - self.total_psm

    def __str__(self) -> str:
        buffer = "key      PSMs   Prime PSM group\n"
        for k, group in self.prime_groups.items():
            buffer += f"{k:>2} {group['count']:>10}   {group['prefixes']}\n"
        buffer += f"\nTotal psm: {self.total_psm}"
        if self.unaccounted_psm:
            buffer += f"\nUnaccounted PSM: {self.unaccounted_psm}"
        return buffer

    def get_prime_groups_for_prefixes(
        self, target_prefixes: list[list[str]]
    ) -> tuple[list[int], list[int]]:
        """
        Return prime PSM groups for the general PSM group defined by its target
        prefixes.

        :param target_prefixes: Collection of prefixes to identify target PSM.
        :return: List of keys for decoy prime groups and list of keys for target groups.
        """
        decoy_keys = []
        target_keys = []
        for i, group in self.prime_groups.items():
            for bunch in target_prefixes:
                if is_derived_superset(
                    group["prefixes"],
                    [self.decoy_prefix + target_prefix for target_prefix in bunch],
                ):
                    decoy_keys.append(i)
                if is_derived_superset(group["prefixes"], [self.decoy_prefix] + bunch):
                    target_keys.append(i)
                if is_derived_superset(group["prefixes"], bunch):
                    target_keys.append(i)
        return decoy_keys, target_keys


class GeneralPsmGroup:
    """General PSM group defined by arbitrary prime groups."""

    def __init__(
        self,
        label: str,
        decoy_keys: list[int],
        target_keys: list[int],
        prime_groups_collection: PrimeGroupsCollection,
    ):
        """
        Group PSMs by arbitrary prime groups.

        :param label: Group label.
        :param decoy_keys: Keys of decoy prime groups.
        :param target_keys: Keys of target prime groups.
        :param prime_groups_collection: Instance of `PrimeGroupsCollection`.
        """
        self.label = label
        self.decoy_keys = decoy_keys
        self.target_keys = target_keys
        self.df_psm = prime_groups_collection.df_psm
        self.prime_groups = prime_groups_collection.prime_groups

        self.decoy_ind = self.prime_groups[self.decoy_keys[0]]["indicator"]
        for i in self.decoy_keys[1:]:
            self.decoy_ind = self.decoy_ind | self.prime_groups[i]["indicator"]

        self.target_ind = self.prime_groups[target_keys[0]]["indicator"]
        for i in target_keys[1:]:
            self.target_ind = self.target_ind | self.prime_groups[i]["indicator"]

        total_decoy_psm = sum(self.prime_groups[i]["count"] for i in self.decoy_keys)
        self.total_target_psm = sum(
            self.prime_groups[i]["count"] for i in self.target_keys
        )
        self.total_psm = total_decoy_psm + self.total_target_psm
        self.percent_factor = 100 / max(self.df_psm.shape[0], 1)
        self.table_header = "key      PSMs         %  Prime PSM group"

    def __str__(self) -> str:
        return self.format(verbose=False)

    def __format_subgroup(self, subgroup_keys: list[int], label: str = None) -> str:
        """Format subgroup for printing."""
        buffer = f"    =={label:^20}==\n" if label else ""
        for i in subgroup_keys:
            count = self.prime_groups[i]["count"]
            percent = self.percent_factor * self.prime_groups[i]["count"]
            buffer += f"{i:>2} {count:>10}   {percent:>7.2f}  {self.prime_groups[i]['prefixes']}\n"
        return buffer[:-1]

    def format(self, verbose=False) -> str:
        """
        Return formatted representation of class instance.

        :param verbose: Show header, decoy subgroup and overall stats.
        :return: Formatted string.
        """
        buffer = self.table_header + "\n"
        if verbose:
            buffer += self.__format_subgroup(self.decoy_keys, "Decoy subgroup")
            buffer += "\n"
            buffer += self.__format_subgroup(self.target_keys, "Target subgroup")
            buffer += "\n\n"
            percent = self.percent_factor * self.total_target_psm
            buffer += (
                f"Total target PSMs: {self.total_target_psm:>10} ({percent:>5.2f}%)\n"
            )
            percent = self.percent_factor * self.total_psm
            buffer += f"Total PSMs:        {self.total_psm:>10} ({percent:>5.2f}%)"
        else:
            buffer += self.__format_subgroup(self.target_keys)
        return buffer


class PsmGroup(GeneralPsmGroup):
    """Group of PSMs defined by collection of its target prefixes."""

    def __init__(
        self,
        label: str,
        target_prefixes: list[list[str]],
        prime_groups_collection: PrimeGroupsCollection,
    ):
        """
        Group PSMs by collection of target prefixes.

        :param label: Group label.
        :param target_prefixes: Collection of target prefixes.
        :param prime_groups_collection: Instance of `PrimeGroupsCollection`.
        """
        self.target_prefixes = target_prefixes
        decoy_keys, target_keys = prime_groups_collection.get_prime_groups_for_prefixes(
            self.target_prefixes
        )
        super().__init__(
            label=label,
            decoy_keys=decoy_keys,
            target_keys=target_keys,
            prime_groups_collection=prime_groups_collection,
        )

    def __str__(self) -> str:
        return self.format(verbose=True)

    def format(self, verbose=False) -> str:
        """
        Return formatted representation of class instance.

        :param verbose: Show header and additional info.
        :return: Formatted string.
        """
        buffer = f"{self.label}: {self.target_prefixes}\n" if verbose else ""
        buffer += super().format(verbose)
        return buffer



"""TMTCrunch main module."""

__all__ = [
    "filter_groupwise",
    "preprocess_peptides",
    "preprocess_psm",
    "process_single_batch",
    "process_fraction",
]


import logging
import os.path
import sys
from argparse import ArgumentParser

import numpy as np
import pandas as pd

from . import __version__
from .altsp import PrimeGroupsCollection, PsmGroup, generate_prefix_collection
from .config import (
    format_settings,
    load_config,
    load_default_config,
    load_phospho_config,
)
from .utils import (
    annotated_int,
    apply_modifications,
    drop_decoys_from_protein_group,
    format_psm_stat,
    generate_psm_stat,
    groupwise_qvalues,
    indicator,
    load_gene_mapping,
    mods_convert_mass_to_label,
    mods_read_from_settings,
    protein_abundance,
    uniq,
)

logger = logging.getLogger(__name__)


def filter_groupwise(df_psm: pd.DataFrame, settings: dict) -> pd.DataFrame:
    """
    Perform qroupwise filtration of PSMs.

    :param df_psm: DataFrame of PSMs (Scavager format)
    :param settings: TMTCrunch settings.
    :return: DataFrame with filtered PSMs.
    """
    decoy_prefix = settings["decoy_prefix"]
    target_prefixes = settings["target_prefixes"]
    requested_groups = settings["psm_group"]

    prefix_collection = generate_prefix_collection(target_prefixes, decoy_prefix)
    primes = PrimeGroupsCollection(prefix_collection, df_psm, decoy_prefix)

    logger.info(f"Prime PSM groups:\n{primes}\n")
    dframes = {group_name: pd.DataFrame() for group_name in requested_groups.keys()}
    for group_name, group_cfg in requested_groups.items():
        group_fdr = group_cfg["fdr"]
        group_psm = PsmGroup(
            group_cfg["descr"],
            target_prefixes=group_cfg["prefixes"],
            prime_groups_collection=primes,
        )
        logger.info(f"{group_psm}\n")
        df_psm_group = groupwise_qvalues(df_psm, group_psm)

        if True:  # display number of passed PSMs for different FDR values.
            fdr_steps = 5
            passed = [
                df_psm_group[df_psm_group["group_q"] < fdr].shape[0]
                for fdr in np.linspace(group_fdr / fdr_steps, group_fdr, fdr_steps)
            ]
            logger.info(f"PSMs at fdr=[{group_fdr / fdr_steps}, {group_fdr}]: {passed}")

        df_psm_group = df_psm_group[df_psm_group["group_q"] < group_fdr]
        df_psm_group = df_psm_group[~df_psm_group["decoy"]]

        if True:  # display groupwise distribution of passed PSMs
            group_psm_passed = PsmGroup(
                f"PSMs passed at fdr={group_fdr}",
                target_prefixes=group_cfg["prefixes"],
                prime_groups_collection=PrimeGroupsCollection(
                    prefix_collection, df_psm_group, decoy_prefix
                ),
            )
            logger.info(
                f"PSMs passed at fdr={group_fdr}: {df_psm_group.shape[0]}\n"
                f"{group_psm_passed.format(False)}\n"
            )

        df_psm_group["psm_group"] = group_name
        dframes[group_name] = df_psm_group
    df_psm = pd.concat(dframes.values(), ignore_index=True)
    return df_psm


def preprocess_peptides(
    df_psm: pd.DataFrame, settings: dict, inplace: bool = True
) -> pd.DataFrame | None:
    """
    Apply modification to peptides.

    :param df_psm: DataFrame of PSMs (Scavager format).
    :param settings: TMTCrunch settings.
    :param inplace: Whether to modify the PSMs DataFrame or create a copy.
    :return: DataFrame with preprocessed peptides or None if `inplace=True`.
    """

    all_mods = settings["all_mods"]
    selective_mods = settings["selective_mods"]

    def convert_all(modifications):
        return mods_convert_mass_to_label(modifications, all_mods, unrecognized=False)

    def convert_selective(modifications):
        return mods_convert_mass_to_label(
            modifications, selective_mods, unrecognized=False
        )

    def find_unknown(modifications):
        return mods_convert_mass_to_label(modifications, all_mods, unrecognized=True)[1]

    def modify_peptide(df):
        return apply_modifications(df.iloc[0], df.iloc[1])

    if not inplace:
        df_psm = df_psm.copy()

    # peptide with selective mods only
    df_psm["modifications_pydict"] = df_psm["modifications"].apply(convert_selective)
    df_psm["modpeptide"] = df_psm[["peptide", "modifications_pydict"]].apply(
        modify_peptide, axis=1
    )
    # peptide with all mods
    df_psm["modifications_pydict"] = df_psm["modifications"].apply(convert_all)
    df_psm["peptide_all_mods"] = df_psm[["peptide", "modifications_pydict"]].apply(
        modify_peptide, axis=1
    )
    # unrecognized modifications
    df_psm["unknown_mods"] = df_psm["modifications"].apply(find_unknown)

    if not inplace:
        return df_psm


def process_fraction(df_psm: pd.DataFrame, settings: dict) -> pd.DataFrame:
    """
    Process PSMs from single fraction.

    :param df_psm: DataFrame of PSMs.
    :param settings: TMTCrunch settings.
    :return: DataFrame with filtered PSMs.
    """
    decoy_prefix = settings["decoy_prefix"]
    with_modifications = settings["with_modifications"]
    global_fdr = settings["global_fdr"]
    groupwise = settings["groupwise"]
    user_keep_cols = settings["keep_columns"]
    tmt_cols = settings["gis_columns"] + settings["specimen_columns"]
    simulate_gis_cols = settings["simulate_gis"]
    with_simulated_gis = len(settings["simulate_gis"])
    fake_gis_column = settings["gis_columns"][0]

    cols = []
    if groupwise:
        df_psm = filter_groupwise(df_psm, settings)
        cols += ["psm_group"]
    else:
        df_psm = df_psm[~df_psm["decoy"]]
        df_psm = df_psm[df_psm["PSM_q"] <= global_fdr]
    if with_modifications:
        preprocess_peptides(df_psm, settings, inplace=True)
        cols += ["modpeptide", "peptide_all_mods", "unknown_mods"]
    if with_simulated_gis:
        df_psm[fake_gis_column] = np.mean(df_psm[simulate_gis_cols], axis=1)

    df_psm["protein"] = df_psm["protein"].apply(
        drop_decoys_from_protein_group, args=(decoy_prefix,)
    )
    df_psm["gene"] = ""
    df_psm["file"] = ""
    cols += [
        "peptide",
        "gene",
        "protein",
        "file",
        "modifications",
        "spectrum",
        "retention_time",
    ]
    cols += list(set(user_keep_cols) - set(cols))
    cols += tmt_cols
    return df_psm[cols]


def preprocess_psm(df_psm: pd.DataFrame, settings: dict) -> tuple[pd.DataFrame, int]:
    """
    Prepare PSMs of a single batch for merging with other batches.

    Reject PSMs with failed channels.
    Normalize tmt channel to account for loading difference.
    Reduce channel intensities with respect to GIS.

    :param df_psm: DataFrame of PSMs.
    :param settings: TMTCrunch settings.
    :return: tuple of DataFrame with preprocessed PSMs and number of rejected PSMs.
    """
    gis_cols = settings["gis_columns"]
    spn_cols = settings["specimen_columns"]
    tmt_cols = settings["gis_columns"] + settings["specimen_columns"]

    # TODO: Drop PSMs only with failed GIS channels.
    # protein_abundance() has to be resistant to the missing values.
    # ind_gis_non_zero = indicator(df_psm, cols=gis_cols, ind_func=bool)
    ind_all_non_zero = indicator(df_psm, cols=tmt_cols, ind_func=bool)
    ind_all_finite = indicator(df_psm, cols=spn_cols, ind_func=np.isfinite)

    n_total = df_psm.shape[0]
    df_psm = df_psm[ind_all_non_zero & ind_all_finite].copy()
    n_rejected = n_total - df_psm.shape[0]

    # Normalize intensity per channel to account for loading difference.
    # If MS/MS were reproducible, sum() could be used for normalization.
    df_psm.loc[:, tmt_cols] /= np.mean(df_psm[tmt_cols], axis=0)
    # Switch to natural logarithm for further analysis.
    # The absolute error for log(x) is the relative error for x
    # due to d(log(x)) = dx/x and we like it.
    df_psm.loc[:, tmt_cols] = np.log(df_psm[tmt_cols])

    df_psm["gis_mean"] = np.mean(df_psm[gis_cols], axis=1)
    df_psm["gis_err"] = np.std(df_psm[gis_cols], axis=1)
    # Reduce individual intensities with respect to the mean GIS intensity.
    df_psm[spn_cols] -= np.array(df_psm["gis_mean"])[:, np.newaxis]
    return df_psm, n_rejected


def process_single_batch(files: list[str], settings: dict) -> dict:
    """
    Process fractions of the same batch.

    Calculate gene product and protein abundance from individual PSMs.
    Group results by PSM groups, genes, and proteins.

    :param files: Scavager *_PSMs_full.tsv files.
    :param settings: TMTCrunch settings.
    :return: dictionary of DataFrames for PSMs, proteins, and genes.
    """
    if settings["input_format"] == "sage":
        from .sage import read_fraction
    else:
        from .scavager import read_fraction

    groupwise = settings["groupwise"]
    requested_groups = settings["psm_group"]
    gis_cols = settings["gis_columns"]
    spn_cols = settings["specimen_columns"]
    with_modifications = settings["with_modifications"]
    output_tables = dict.fromkeys(settings["output_tables"])
    fasta = settings["fasta_file"]

    logger.info(f"Total files in the batch: {len(files)}")
    genes = load_gene_mapping(fasta) if fasta != "" else {}
    fractions = []
    for file in files:
        logger.info(f"Processing {file}")
        df = process_fraction(read_fraction(file), settings)
        df["file"] = file
        fractions.append(df)
    df_psm = pd.concat(fractions, ignore_index=True)
    del fractions, df

    psm_stat = [annotated_int("Total PSMs", df_psm.shape[0])]
    df_psm, n_psm_bad = preprocess_psm(df_psm, settings)
    psm_stat.append(annotated_int("PSMs with failed channels", n_psm_bad))
    psm_stat.append(annotated_int("PSMs used for assembling", df_psm.shape[0]))
    logger.info("Summary:\n" + format_psm_stat(psm_stat))

    df_psm["gene"] = df_psm["protein"].apply(
        lambda proteins: uniq([genes.get(p, "_not_found") for p in proteins])
    )
    df_psm["gene"] = df_psm["gene"].apply(";".join)
    df_psm["protein"] = df_psm["protein"].apply(";".join)

    # dictionary {column: description}
    stat_descr = {
        "peptide_all_mods": "Unique peptides (all modifications)",
        "modpeptide": "Unique peptides (selective modifications)",
        "peptide": "Unique peptides (base sequence)",
        "protein": "Protein groups",
        "gene": "Gene groups",
    }
    selected_descr = {
        "modpeptide": "Unique peptides",
        "peptide": "Unique peptides (base sequence)",
        "protein": "Protein groups",
        "gene": "Gene groups",
    }
    psm_stat = generate_psm_stat(df_psm, stat_descr)
    logger.info("Summary for all PSMs:\n" + format_psm_stat(psm_stat))

    if groupwise:
        for group, group_cfg in requested_groups.items():
            df = df_psm[df_psm["psm_group"] == group]
            psm_stat = generate_psm_stat(df, stat_descr)
            group_descr = group_cfg["descr"]
            summary = f"Summary for {group_descr}:\n" + format_psm_stat(psm_stat)
            if with_modifications:
                ind = df["modpeptide"] != df["peptide"]
                selected_stat = generate_psm_stat(df[ind], selected_descr)
                summary += "\tWith selective modifications:\n"
                summary += format_psm_stat(selected_stat)
            logger.info(summary)

    # Create gis table
    if "gis" in output_tables.keys():
        output_gis_index = ["file"]
        output_gis_index += ["psm_group"] if groupwise else []
        output_gis_index += ["gene", "protein", "peptide"]
        if with_modifications:
            output_gis_index += ["modpeptide", "peptide_all_mods"]
        output_gis_columns = ["spectrum", "retention_time"] + gis_cols
        output_tables["gis"] = (
            df_psm[output_gis_index + output_gis_columns]
            .copy()
            .sort_values(by="retention_time")
        )

    # Group PSMs with the same gene, protein, peptide, peptide with mods to calculate corresponding abundance.
    supported_output_tables = ["gene", "protein", "peptide", "modpeptide"]
    index_columns = ["psm_group"] if groupwise else []
    groupby_cols = {}
    for kind in supported_output_tables:
        index_columns += [kind]
        groupby_cols[kind] = index_columns.copy()

    finest_level = "modpeptide" if with_modifications else "peptide"
    df_psm.sort_values(by=groupby_cols[finest_level], inplace=True)
    df_psm_short = df_psm[groupby_cols[finest_level] + spn_cols].copy()
    # GIS error is the only source of error we currently account for specimen values at PSMs level.
    # Since we work with natural log(intensity) the specimen error is equal to GIS error.
    # The GIS error is undefined for data with only one GIS channel.
    spn_err_cols = None
    if len(gis_cols) >= 2:
        spn_err_cols = [f"{col}_err" for col in spn_cols]
        df_psm_short[spn_err_cols] = (
            np.ones(df_psm[spn_cols].shape) * np.array(df_psm["gis_err"])[:, np.newaxis]
        )
    # Calculate abundance at gene, protein, peptide, etc levels.
    for kind in groupby_cols.keys():
        if kind in output_tables.keys():
            logger.info(f"Calculating abundance at {kind} level.")
            output_tables[kind] = protein_abundance(
                df_psm_short,
                groupby_cols[kind],
                spn_cols,
                spn_err_cols,
            )

    if "psm" in output_tables.keys():
        output_tables["psm"] = df_psm
    return output_tables


def cli_main() -> None:
    parser = ArgumentParser(description=f"TMTCrunch version {__version__}")
    supported_input = ["auto", "scavager", "sage"]
    parser.add_argument(
        "fractions",
        nargs="*",
        help="Scavager *_PSMs_full.tsv files or directories with Sage search results.",
    )
    parser.add_argument(
        "--cfg",
        action="append",
        help="Path to configuration file. Can be specified multiple times.",
    )
    parser.add_argument(
        "--fasta",
        help="Path to protein fasta file for mapping protein to gene symbol.",
    )
    parser.add_argument(
        "--input-format",
        type=str,
        choices=supported_input,
        default="auto",
        help="Format of input data. Supported: "
        + ", ".join([f"'{s}'" for s in supported_input])
        + ". Default is 'auto'",
    )
    parser.add_argument(
        "--output-dir",
        "--odir",
        default="",
        help="Existing output directory. Default is current directory.",
    )
    parser.add_argument(
        "--output-prefix",
        "--oprefix",
        default="tmtcrunch_",
        help="Prefix for output files. Default is 'tmtcrunch_'.",
    )
    parser.add_argument(
        "--phospho",
        action="store_true",
        help="Enable common modifications for phospho-proteomics.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        choices=range(3),
        default=1,
        help="Logging verbosity. Default is 1.",
    )
    parser.add_argument(
        "--show-config", action="store_true", help="Show configuration and exit."
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"{__version__}",
        help="Output version information and exit.",
    )

    cmd_args = parser.parse_args()
    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    logging.basicConfig(
        format="{levelname}: {message}",
        datefmt="[%H:%M:%S]",
        level=log_levels[cmd_args.verbose],
        style="{",
    )

    with_gene_table = False
    settings = load_default_config()
    if cmd_args.phospho:
        settings |= load_phospho_config()
    # Supported tables: gene, protein, peptide, modpeptide, psm, gis.
    settings["output_tables"] = ["gis", "psm"]
    if cmd_args.cfg:
        for fpath in cmd_args.cfg:
            settings |= load_config(fpath)

    if cmd_args.fasta:
        settings["fasta_file"] = cmd_args.fasta

    if settings["fasta_file"] != "":
        with_gene_table = True

    if not (len(settings["gis_columns"]) or len(settings["simulate_gis"])):
        logger.error("At least one GIS column is required!")
        sys.exit(1)
    conflicting = set(settings["gis_columns"]) & set(settings["specimen_columns"])
    if conflicting:
        logger.error(f"Overlapping GIS and specimen columns: {conflicting}")
        sys.exit(1)

    # prepare for groupwise
    if settings["groupwise"]:
        for group_name, group_cfg in settings["psm_group"].items():
            if "fdr" not in group_cfg.keys():
                settings[group_name]["fdr"] = settings["global_fdr"]
    else:
        settings["psm_group"] = {}

    if settings["groupwise"] and "target_prefixes" not in settings.keys():
        target_prefixes = []
        for group_cfg in settings["psm_group"].values():
            for prefixes in group_cfg["prefixes"]:
                target_prefixes.extend(prefixes)
        settings["target_prefixes"] = uniq(target_prefixes)

    # prepare for modifications
    if settings["with_modifications"]:
        selective_mods = mods_read_from_settings(settings["modification"]["selective"])
        universal_mods = mods_read_from_settings(settings["modification"]["universal"])
        settings["selective_mods"] = selective_mods
        settings["all_mods"] = selective_mods | universal_mods
        settings["output_tables"] += ["modpeptide"]
    else:
        settings["output_tables"] += ["peptide", "protein"]
        if with_gene_table:
            settings["output_tables"] += ["gene"]

    settings["input_format"] = cmd_args.input_format
    if len(cmd_args.fractions) > 0 and cmd_args.input_format == "auto":
        is_tsv = [f.endswith("PSMs_full.tsv") for f in cmd_args.fractions]
        if all(is_tsv):
            settings["input_format"] = "scavager"
        elif not any(is_tsv):
            settings["input_format"] = "sage"
        else:
            logger.error(f"failed to detect input format: {cmd_args.fractions}")
            sys.exit(1)

    if cmd_args.show_config:
        print(f"output directory: {cmd_args.output_dir}")
        for kind in settings["output_tables"]:
            print(f"output file: {cmd_args.output_prefix}{kind}.tsv")
        print(f"verbosity: {cmd_args.verbose}")
        print(format_settings(settings))
        sys.exit()
    if len(cmd_args.fractions) == 0:
        logger.error(f"missing argument: file")
        sys.exit(1)

    logger.info(
        "Starting...\n"
        + f"TMTCrunch version {__version__}\n"
        + format_settings(settings)
    )
    if len(settings["gis_columns"]) == 1:
        logger.warning(
            "Only one GIS channel is specified. Using simplified quantification."
        )
    if len(settings["simulate_gis"]):
        settings["gis_columns"] = ["fake_gis"]

    output_tables = process_single_batch(cmd_args.fractions, settings)
    for kind, df in output_tables.items():
        fpath = os.path.join(cmd_args.output_dir, f"{cmd_args.output_prefix}{kind}.tsv")
        logger.info(f"Saving {fpath}")
        df.to_csv(fpath, sep="\t", index=False)
    logger.info("Done.")


if __name__ == "__main__":
    cli_main()



"""Miscellaneous functions."""

__all__ = [
    "standard_residue_mass",
    "protein_abundance",
    "drop_decoys_from_protein_group",
    "get_gene_name",
    "load_gene_mapping",
    "groupwise_qvalues",
    "indicator",
    "mod_mass_repr",
    "mods_read_from_settings",
    "mods_convert_mass_to_label",
    "apply_modifications",
    "weighted_mean",
    "mean_and_std",
    "uniq",
    "annotated_int",
    "generate_psm_stat",
    "format_psm_stat",
]


import logging
import re
from collections import namedtuple
from collections.abc import Iterable
from typing import Callable, Union

import numpy as np

try:
    from astropy.stats import biweight_scale, median_absolute_deviation
except ModuleNotFoundError:
    from .astropy_stats import biweight_scale, median_absolute_deviation

import pandas as pd
from pyteomics import auxiliary as paux
from pyteomics.mass import std_aa_mass

from .altsp import GeneralPsmGroup

logger = logging.getLogger(__name__)


standard_residue_mass = std_aa_mass | {"-": 1.00782503207}


def uniq(x: Iterable[str], sort=True) -> list:
    """
    Filter out repeated elements from the input list.

    :param x: Iterable.
    :param sort: If True, return sorted list.
    :return: List of unique elements.
    """
    y = list(set(x))
    y = sorted(y) if sort else y
    return y


def get_gene_name(protein_descr: str, missing_gene: str = "_not_found") -> str:
    """
    Parse protein description for gene name. Supported fasta format:
    UniProt/Swiss-Prot, Ensembl.

    :param protein_descr: Protein description from FASTA header.
    :param missing_gene: String to return if gene name is not found.
    :return: Gene name.
    """
    pattern = r"(GN=|gene_symbol:)(\S*)"
    res = re.search(pattern, protein_descr)
    if res:
        return res.group(2)
    return missing_gene


def load_gene_mapping(fpath: str) -> dict:
    """
    Read protein fasta and return {"protein id": "gene_name"} dict.

    :param fpath: path to fasta file.
    :return: dict of {protein: gene}.
    """
    mapping = {}
    missing_gene = "_not_found"
    with open(fpath, "rt") as f:
        for line in f:
            if line[0] == ">":
                res = line[1:].rstrip().split(sep=" ", maxsplit=1)
                protein = res[0]
                descr = res[1] if len(res) > 1 else ""
                gene = get_gene_name(descr, missing_gene)
                mapping[protein] = gene
    genes = uniq(mapping.values())
    total = len(genes)
    if len(genes) == 0 or len(genes) == 1 and genes[0] == "_not_found":
        logger.warning("No gene info in protein fasta headers.")
    else:
        logger.info(f"Total gene names found in protein fasta headers: {total}")
    return mapping


def drop_decoys_from_protein_group(protein_group: list, decoy_prefix: str = "DECOY_"):
    """
    Drop decoy proteins from protein group and sort remaining proteins.

    :param protein_group: list of proteins
    :param decoy_prefix: Decoy prefix, defaults to 'DECOY_'.
    :return: sorted list of proteins
    """
    # work around Identipy erroneously reporting duplicates in protein group
    # return sorted([p for p in protein_group if not p.startswith(decoy_prefix)])
    return uniq([p for p in protein_group if not p.startswith(decoy_prefix)])


def groupwise_qvalues(
    df_psm: pd.DataFrame,
    psm_group: GeneralPsmGroup,
    training: str = "decoy_training",
    testing: str = "decoy_testing",
    score: str = "PSM_score",
) -> pd.DataFrame:
    """
    Calculate groupwise qvalues for `psm_group` in `df_psm`. Return new DataFrame with
    group PSMs and q-values stored in `group_q` column.

    :param df_psm: Scavager PSMs_full DataFrame.
    :param psm_group: PSMs group to calculate qvalues.
    :param training: column name for training decoys.
    :param testing: column name for testing decoys.
    :param score: column name for PSM score.
    :return: DataFrame with group PSMs.
    """

    group_ind = psm_group.target_ind | psm_group.decoy_ind
    df_group = df_psm[group_ind]

    group_pep_ratio = df_group[testing].sum() / df_group["decoy"].sum()
    # filter out decoy which were used for CatBoost training
    df_group = df_group[~df_group[training]].copy()

    res = paux.qvalues(
        df_group,
        key=score,
        is_decoy=testing,
        remove_decoy=False,
        ratio=group_pep_ratio,
        formula=1,
        full_output=True,
        correction=1,
    )
    df_group["group_q"] = res["q"]
    return df_group


def indicator(
    df: pd.DataFrame, cols: list[str], ind_func: Callable[[float], bool] = bool
) -> pd.Series:
    """
    Return boolean array for indexing rows where `ind_func` is True for all columns in
    the DataFrame.

    :param df: DataFrame.
    :param cols: List of columns.
    :param ind_func: Boolean function.
    :return: Boolean array.
    """
    return df[cols].map(ind_func).all(axis="columns")


def mod_mass_repr(mass) -> str:
    """
    String representation of modification mass, rounded to 3 decimal places.
    To be used as a key for modification dict.

    :param mass: mass of modification.
    :return: string representation.
    """
    return format(float(mass), "0.3f")


def mods_read_from_settings(modifications: dict, residue_mass: dict = None) -> dict:
    """
    Convert modifications from TMTCrunch settings to the pairs of total mass
    of modified amino acid residue and modX label.

    :param modifications: dict of modifications from TMTCrunch settings.
    :param residue_mass: animo acid residue masses. Default is `standard_residue_mass`.
    :return: dict of {"mass": "label"}
    """
    if residue_mass is None:
        residue_mass = standard_residue_mass
    mods = {}
    for mod in modifications.values():
        for site in mod["site"].replace("^", "-"):
            mass = residue_mass[site] + mod["mass_delta"]
            modX = mod["modX"] + site
            mods[mod_mass_repr(mass)] = modX
    return mods


def mods_convert_mass_to_label(
    modifications: dict, label_for_mass: dict, unrecognized: bool = True
) -> dict | tuple[dict, dict]:
    """
    Convert modifications from {position: "mass"} to {position: "label"} format using
    dict of {"mass": "label"} and return converted modifications and optionally
    unrecognized modifications.

    :param modifications: dict of modifications {position: "mass"}.
    :param label_for_mass: dict of {"mass": "label"}.
    :param unrecognized: If True, return also a dict of unrecognized modifications.
    :return: dict of modifications or tuple of dict of modifications and dict of
        unrecognized modifications.
    """
    recognized_mods = {
        i: label_for_mass[mass]
        for i, mass in modifications.items()
        if mass in label_for_mass.keys()
    }
    if unrecognized:
        unknown_mods_loc = list(modifications.keys() - recognized_mods.keys())
        unknown_mods = {i: modifications[i] for i in unknown_mods_loc}
        return recognized_mods, unknown_mods
    else:
        return recognized_mods


def apply_modifications(sequence: str, modifications: dict) -> str:
    """
    Apply modifications for amino- sequence and return modified sequence.

    :param sequence: Original peptide sequence.
    :param modifications: dict of modifications {location: "label"}.
    :return: Modified peptide sequence.
    """
    if len(modifications) == 0:
        return sequence

    modified = ""
    # Note: First aminoacid in a sequence has location 1.
    # Modification at zero location is prepended to the sequence.
    mod_loc = sorted(modifications.keys())

    # Prepend n-term modification.
    if mod_loc[0] == 0:
        modified = modifications[0]
        if len(mod_loc) == 1:  # n-term was the only modification.
            modified += sequence
            return modified
        else:
            mod_loc = mod_loc[1:]

    # Transform sequence.
    for i_l, i_r in zip([0] + mod_loc[:-1], mod_loc):
        modified += sequence[i_l : i_r - 1] + modifications[i_r]
    # Append sequence tail.
    modified += sequence[mod_loc[-1] :]

    return modified


def weighted_mean(
    data: np.array,
    data_err: np.array = None,
    axis: int = 0,
    c: float = 6.0,
    mean0=None,
) -> Union[float, np.array]:
    """
    Calculate the weighted mean along the specified axis.

    If `data_err` is not None, it is added to the weights and then biweight location
    is computed. Otherwise, the weighted mean is equivalent to the biweight_location.
    Based on biweight location:
    https://docs.astropy.org/en/stable/api/astropy.stats.biweight.biweight_location.html

    :param data: Input data.
    :param data_err: Data uncertainty, optional.
    :param axis: Axis along which a mean is calculated.
    :param c: Tuning constant for the biweight estimator.
    :param mean0: Initial guess for mean.
    :return: Weighted mean of the input data.
    """
    if mean0 is None:
        mean0 = np.median(data, axis=axis)

    mad = median_absolute_deviation(data, axis=axis)

    d = data - mean0
    weight = d**2
    if data_err is not None:
        weight += data_err**2

    if np.isscalar(mean0) and (mad == 0.0 or np.isnan(mad)):
        correction = 0
    else:
        with np.errstate(divide="ignore", invalid="ignore"):
            u = weight / (c * mad) ** 2
            mask = u >= 1
            u = (1 - u) ** 2
            u[mask] = 0
            correction = np.sum(u * d, axis=axis) / np.sum(u, axis=axis)

    if np.isscalar(mean0):
        # if data_err is big, sum(u) could be 0 and correction is undefined
        correction = 0 if np.isnan(correction) else correction
        wmean = mean0 + correction
    else:
        correction = np.where(np.isnan(correction), 0, correction)
        wmean = mean0 + np.where(mad == 0, 0, correction)

    return wmean


def mean_and_std(
    data: np.array, data_err: np.array, axis: int = 0
) -> tuple[np.array, np.array]:
    """
    Robust estimation for the mean and std based on Tukey's biweight.

    :param data: Input data.
    :param data_err: Data uncertainty, optional.
    :param axis: Axis along which statistic is calculated.
    :return: Tuple of mean and standard deviation.
    """
    wmean0 = weighted_mean(data, data_err, axis=axis)
    wmean = weighted_mean(data, data_err, axis=axis, mean0=wmean0)
    std = biweight_scale(data, axis=axis, M=wmean)
    return wmean, std


def protein_abundance(
    df_psm: pd.DataFrame,
    groupby_cols: list,
    specimen_cols: list,
    error_cols: list = None,
) -> pd.DataFrame:
    """
    Calculate protein abundance and its standard deviation.

    The protein abundance is computed as a weighted mean of reduced intensities of PSMs.
    Standard deviation of abundance is computed by biweight_scale.

    :param df_psm: DataFrame with reduced intensities and optionally errors.
    :param groupby_cols: df_psm columns to group PSMs for a protein.
    :param specimen_cols: List of specimen columns.
    :param error_cols: List of error columns corresponding to the specimen columns,
        optional.
    :return: DataFrame of protein abundance and its std error.
    """
    if len(groupby_cols) > 1:
        multi_index = pd.MultiIndex.from_tuples(
            df_psm.groupby(groupby_cols).indices.keys(),
            names=groupby_cols,
        )
    elif len(groupby_cols) == 1:
        multi_index = pd.Index(
            df_psm.groupby(groupby_cols).indices.keys(),
            name=groupby_cols[0],
        )
    else:
        raise ValueError("`groupby_cols` must be non-empty list.")
    std_cols = [f"{col}_std" for col in specimen_cols]
    df_protein = pd.DataFrame(
        index=multi_index,
        columns=specimen_cols + std_cols,
        dtype=float,
    )
    for mi, df in df_psm.groupby(groupby_cols):
        data = np.array(df[specimen_cols])
        data_err = None if error_cols is None else np.array(df[error_cols])
        df_protein.loc[mi] = np.hstack(mean_and_std(data, data_err))
    return df_protein.reset_index()


annotated_int = namedtuple("annotated_int", ["descr", "value"])
"""Tuple for integer value and its description."""


def generate_psm_stat(df: pd.DataFrame, stat_descr: dict) -> list[annotated_int]:
    """
    Calculate number of unique values in columns.

    :param df: DataFrame of PSMs.
    :param stat_descr: Dict {"column": "description"}.
    :return: List of annotated_int.
    """
    psm_stat = []
    for col, descr in stat_descr.items():
        if col in df.columns:
            n = len(uniq(df[col], sort=False))
            psm_stat.append(annotated_int(descr, n))
    return psm_stat


def format_psm_stat(psm_stat: list[annotated_int], width: tuple[int] = (50, 10)) -> str:
    """
    Return formatted representation for a list of annotated values.

    :param psm_stat: List of annotated values.
    :param width: Tuple of field widths.
    :return: Formatted string.
    """
    stat_str = ""
    for record in psm_stat:
        stat_str += (
            format(f"{record.descr}:", f"<{width[0]}")
            + format(f"{record.value}", f">{width[1]}")
            + "\n"
        )
    return stat_str



"""Scavager helpers."""

__all__ = [
    "read_fraction",
]


from ast import literal_eval

import pandas as pd

from .utils import mod_mass_repr


def read_fraction(fpath: str) -> pd.DataFrame:
    """
    Read and reformat Scavager PSMs_full.tsv file.

    :param fpath: path to the Scavager file.
    :return: DataFrame
    """
    eval_cols = ["protein", "modifications"]
    df_psm = pd.read_table(fpath, converters={key: literal_eval for key in eval_cols})

    mapping = {
        "ML score": "PSM_score",
        "RT exp": "retention_time",
        "decoy1": "decoy_training",
        "decoy2": "decoy_testing",
        "q": "PSM_q",
    }
    df_psm.rename(columns=mapping, inplace=True)

    df_psm["file"] = fpath
    df_psm["modifications"] = df_psm["modifications"].apply(mods_from_scavager)

    return df_psm


def mods_from_scavager(scavager_modifications: list) -> dict:
    """
    Convert list of modifications from the Scavager format ["location@total_mass"] to
    the dict {location: "total_mass"}.

    :param scavager_modifications: list of modifications.
    :return: dict of modifications.
    """
    mods = [mod.split("@") for mod in scavager_modifications]
    return {int(loc): mod_mass_repr(mass) for mass, loc in mods}



"""Sage helpers."""

__all__ = [
    "read_fraction",
]


import os.path
import re

import pandas as pd

from .utils import mod_mass_repr, standard_residue_mass


def read_fraction(fpath: str) -> pd.DataFrame:
    """
    Read and reformat Sage output files.

    :param fpath: path to the Sage output directory.
    :return: DataFrame
    """
    search_file = os.path.join(fpath, "results.sage.tsv")
    tmt_file = os.path.join(fpath, "tmt.tsv")
    df_search = pd.read_table(search_file, index_col=["filename", "scannr"])
    df_tmt = pd.read_table(tmt_file, index_col=["filename", "scannr"])
    df_psm = pd.concat([df_search, df_tmt], axis=1, join="inner")
    df_psm.reset_index(inplace=True)

    mapping = {
        "proteins": "protein",
        "rt": "retention_time",
        "sage_discriminant_score": "PSM_score",
        "scannr": "spectrum",
        "spectrum_q": "PSM_q",
    }
    df_psm.rename(columns=mapping, inplace=True)

    df_psm["decoy"] = df_psm["label"].apply(lambda x: True if x == -1 else False)
    df_psm["decoy_training"] = False
    df_psm["decoy_testing"] = df_psm["decoy"]
    df_psm["file"] = fpath
    df_psm["peptide_mod_delta"] = df_psm["peptide"]  # peptide with modifications
    df_psm["peptide"] = df_psm["peptide"].apply(strip_modifications)
    df_psm["mod_delta"] = df_psm["peptide_mod_delta"].apply(extract_modifications)
    df_psm["modifications"] = df_psm[["peptide", "mod_delta"]].apply(
        lambda df: mods_total_mass(df.iloc[0], df.iloc[1]), axis=1
    )
    df_psm["protein"] = df_psm["protein"].apply(lambda x: x.split(";"))
    df_psm["PSM_score"] = -df_psm["PSM_score"]  # less is better

    return df_psm


def strip_modifications(sequence: str) -> str:
    """
    Strip modifications from peptide sequence.

    :param sequence: peptide sequence with modifications.
    :return: peptide sequence.
    """
    res = re.findall(r"[A-Z]+", sequence)
    peptide = "".join(res)
    return peptide


def extract_modifications(sequence: str) -> dict:
    """
    Extract modifications from peptide sequence to dict
    {position: "mass"}.

    :param sequence: peptide sequence with modifications.
    :return: dict of modifications.
    """
    pattern = r"\[(\+\d+\.\d+)\]\-?"
    remainder = sequence
    modifications = {}
    position = 0
    while True:
        match = re.search(pattern, remainder)
        if match:
            position += match.start()
            modifications[position] = match.group(1)
            remainder = remainder[match.end() :]
        else:
            break
    return modifications


def mods_total_mass(peptide: str, modifications: dict, residue_mass: dict = None):
    """
    Replace delta mass of modification with the total mass of the site and modification.

    :param peptide: unmodified peptide sequence.
    :param modifications: dict of modifications {position: "delta mass"}.
    :param residue_mass: animo acid residue masses. Default is `standard_residue_mass`.
    :return: dict of modifications {position: "total mass"}.
    """
    if residue_mass is None:
        residue_mass = standard_residue_mass
    peptide = "-" + peptide
    return {
        pos: mod_mass_repr(residue_mass[peptide[pos]] + float(mass_delta))
        for pos, mass_delta in modifications.items()
    }



# Licensed under a 3-clause BSD style license
"""
Several functions from astropy.stats in case astropy package is not installed.

https://www.astropy.org/
"""

import numpy as np

__all__ = [
    "median_absolute_deviation",
    "biweight_scale",
]


def median_absolute_deviation(data, axis=None, func=None, ignore_nan=False):
    """
    Calculate the median absolute deviation (MAD).

    The MAD is defined as ``median(abs(a - median(a)))``.

    Parameters
    ----------
    data : array-like
        Input array or object that can be converted to an array.
    axis : None, int, or tuple of int, optional
        The axis or axes along which the MADs are computed.  The default
        (`None`) is to compute the MAD of the flattened array.
    func : callable, optional
        The function used to compute the median. Defaults to `numpy.ma.median`
        for masked arrays, otherwise to `numpy.median`.
    ignore_nan : bool
        Ignore NaN values (treat them as if they are not in the array) when
        computing the median.  This will use `numpy.ma.median` if ``axis`` is
        specified, or `numpy.nanmedian` if ``axis==None`` and numpy's version
        is >1.10 because nanmedian is slightly faster in this case.

    Returns
    -------
    mad : float or `~numpy.ndarray`
        The median absolute deviation of the input array.  If ``axis``
        is `None` then a scalar will be returned, otherwise a
        `~numpy.ndarray` will be returned.
    """
    if func is None:
        # Check if the array has a mask and if so use np.ma.median
        # See https://github.com/numpy/numpy/issues/7330 why using np.ma.median
        # for normal arrays should not be done (summary: np.ma.median always
        # returns an masked array even if the result should be scalar). (#4658)
        if isinstance(data, np.ma.MaskedArray):
            is_masked = True
            func = np.ma.median
            if ignore_nan:
                data = np.ma.masked_where(np.isnan(data), data, copy=True)
        elif ignore_nan:
            is_masked = False
            func = np.nanmedian
        else:
            is_masked = False
            func = np.median  # drops units if result is NaN
    else:
        is_masked = None

    data = np.asanyarray(data)
    # np.nanmedian has `keepdims`, which is a good option if we're not allowing
    # user-passed functions here
    data_median = func(data, axis=axis)

    # broadcast the median array before subtraction
    if axis is not None:
        data_median = np.expand_dims(data_median, axis=axis)

    result = func(np.abs(data - data_median), axis=axis, overwrite_input=True)

    if axis is None and np.ma.isMaskedArray(result):
        # return scalar version
        result = result.item()
    elif np.ma.isMaskedArray(result) and not is_masked:
        # if the input array was not a masked array, we don't want to return a
        # masked array
        result = result.filled(fill_value=np.nan)

    return result


def _stat_functions(data, ignore_nan=False):
    if isinstance(data, np.ma.MaskedArray):
        median_func = np.ma.median
        sum_func = np.ma.sum
    elif ignore_nan:
        median_func = np.nanmedian
        sum_func = np.nansum
    else:
        median_func = np.median
        sum_func = np.sum

    return median_func, sum_func


def biweight_scale(
    data, c=9.0, M=None, axis=None, modify_sample_size=False, *, ignore_nan=False
):
    r"""
    Compute the biweight scale.

    The biweight scale is a robust statistic for determining the
    standard deviation of a distribution.  It is the square root of the
    `biweight midvariance
    <https://en.wikipedia.org/wiki/Robust_measures_of_scale#The_biweight_midvariance>`_.
    It is given by:

    .. math::

        \zeta_{biscl} = \sqrt{n} \ \frac{\sqrt{\sum_{|u_i| < 1} \
            (x_i - M)^2 (1 - u_i^2)^4}} {|(\sum_{|u_i| < 1} \
            (1 - u_i^2) (1 - 5u_i^2))|}

    where :math:`x` is the input data, :math:`M` is the sample median
    (or the input location) and :math:`u_i` is given by:

    .. math::

        u_{i} = \frac{(x_i - M)}{c * MAD}

    where :math:`c` is the tuning constant and :math:`MAD` is the
    `median absolute deviation
    <https://en.wikipedia.org/wiki/Median_absolute_deviation>`_.  The
    biweight midvariance tuning constant ``c`` is typically 9.0 (the
    default).

    If :math:`MAD` is zero, then zero will be returned.

    For the standard definition of biweight scale, :math:`n` is the
    total number of points in the array (or along the input ``axis``, if
    specified).  That definition is used if ``modify_sample_size`` is
    `False`, which is the default.

    However, if ``modify_sample_size = True``, then :math:`n` is the
    number of points for which :math:`|u_i| < 1` (i.e. the total number
    of non-rejected values), i.e.

    .. math::

        n = \sum_{|u_i| < 1} \ 1

    which results in a value closer to the true standard deviation for
    small sample sizes or for a large number of rejected values.

    Parameters
    ----------
    data : array-like
        Input array or object that can be converted to an array.
        ``data`` can be a `~numpy.ma.MaskedArray`.
    c : float, optional
        Tuning constant for the biweight estimator (default = 9.0).
    M : float or array-like, optional
        The location estimate.  If ``M`` is a scalar value, then its
        value will be used for the entire array (or along each ``axis``,
        if specified).  If ``M`` is an array, then its must be an array
        containing the location estimate along each ``axis`` of the
        input array.  If `None` (default), then the median of the input
        array will be used (or along each ``axis``, if specified).
    axis : None, int, or tuple of int, optional
        The axis or axes along which the biweight scales are computed.
        If `None` (default), then the biweight scale of the flattened
        input array will be computed.
    modify_sample_size : bool, optional
        If `False` (default), then the sample size used is the total
        number of elements in the array (or along the input ``axis``, if
        specified), which follows the standard definition of biweight
        scale.  If `True`, then the sample size is reduced to correct
        for any rejected values (i.e. the sample size used includes only
        the non-rejected values), which results in a value closer to the
        true standard deviation for small sample sizes or for a large
        number of rejected values.
    ignore_nan : bool, optional
        Whether to ignore NaN values in the input ``data``.

    Returns
    -------
    biweight_scale : float or `~numpy.ndarray`
        The biweight scale of the input data.  If ``axis`` is `None`
        then a scalar will be returned, otherwise a `~numpy.ndarray`
        will be returned.

    See Also
    --------
    biweight_midvariance, biweight_midcovariance, biweight_location, astropy.stats.mad_std, astropy.stats.median_absolute_deviation

    References
    ----------
    .. [1] Beers, Flynn, and Gebhardt (1990; AJ 100, 32) (https://ui.adsabs.harvard.edu/abs/1990AJ....100...32B)

    .. [2] https://www.itl.nist.gov/div898/software/dataplot/refman2/auxillar/biwscale.htm

    """
    return np.sqrt(
        biweight_midvariance(
            data,
            c=c,
            M=M,
            axis=axis,
            modify_sample_size=modify_sample_size,
            ignore_nan=ignore_nan,
        )
    )


def biweight_midvariance(
    data, c=9.0, M=None, axis=None, modify_sample_size=False, *, ignore_nan=False
):
    r"""
    Compute the biweight midvariance.

    The biweight midvariance is a robust statistic for determining the
    variance of a distribution.  Its square root is a robust estimator
    of scale (i.e. standard deviation).  It is given by:

    .. math::

        \zeta_{bivar} = n \ \frac{\sum_{|u_i| < 1} \
            (x_i - M)^2 (1 - u_i^2)^4} {(\sum_{|u_i| < 1} \
            (1 - u_i^2) (1 - 5u_i^2))^2}

    where :math:`x` is the input data, :math:`M` is the sample median
    (or the input location) and :math:`u_i` is given by:

    .. math::

        u_{i} = \frac{(x_i - M)}{c * MAD}

    where :math:`c` is the tuning constant and :math:`MAD` is the
    `median absolute deviation
    <https://en.wikipedia.org/wiki/Median_absolute_deviation>`_.  The
    biweight midvariance tuning constant ``c`` is typically 9.0 (the
    default).

    If :math:`MAD` is zero, then zero will be returned.

    For the standard definition of `biweight midvariance
    <https://en.wikipedia.org/wiki/Robust_measures_of_scale#The_biweight_midvariance>`_,
    :math:`n` is the total number of points in the array (or along the
    input ``axis``, if specified).  That definition is used if
    ``modify_sample_size`` is `False`, which is the default.

    However, if ``modify_sample_size = True``, then :math:`n` is the
    number of points for which :math:`|u_i| < 1` (i.e. the total number
    of non-rejected values), i.e.

    .. math::

        n = \sum_{|u_i| < 1} \ 1

    which results in a value closer to the true variance for small
    sample sizes or for a large number of rejected values.

    Parameters
    ----------
    data : array-like
        Input array or object that can be converted to an array.
        ``data`` can be a `~numpy.ma.MaskedArray`.
    c : float, optional
        Tuning constant for the biweight estimator (default = 9.0).
    M : float or array-like, optional
        The location estimate.  If ``M`` is a scalar value, then its
        value will be used for the entire array (or along each ``axis``,
        if specified).  If ``M`` is an array, then its must be an array
        containing the location estimate along each ``axis`` of the
        input array.  If `None` (default), then the median of the input
        array will be used (or along each ``axis``, if specified).
    axis : None, int, or tuple of int, optional
        The axis or axes along which the biweight midvariances are
        computed.  If `None` (default), then the biweight midvariance of
        the flattened input array will be computed.
    modify_sample_size : bool, optional
        If `False` (default), then the sample size used is the total
        number of elements in the array (or along the input ``axis``, if
        specified), which follows the standard definition of biweight
        midvariance.  If `True`, then the sample size is reduced to
        correct for any rejected values (i.e. the sample size used
        includes only the non-rejected values), which results in a value
        closer to the true variance for small sample sizes or for a
        large number of rejected values.
    ignore_nan : bool, optional
        Whether to ignore NaN values in the input ``data``.

    Returns
    -------
    biweight_midvariance : float or `~numpy.ndarray`
        The biweight midvariance of the input data.  If ``axis`` is
        `None` then a scalar will be returned, otherwise a
        `~numpy.ndarray` will be returned.

    See Also
    --------
    biweight_midcovariance, biweight_midcorrelation, astropy.stats.mad_std, astropy.stats.median_absolute_deviation

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Robust_measures_of_scale#The_biweight_midvariance

    .. [2] Beers, Flynn, and Gebhardt (1990; AJ 100, 32) (https://ui.adsabs.harvard.edu/abs/1990AJ....100...32B)

    """
    median_func, sum_func = _stat_functions(data, ignore_nan=ignore_nan)

    if isinstance(data, np.ma.MaskedArray) and ignore_nan:
        data = np.ma.masked_where(np.isnan(data), data, copy=True)

    data = np.asanyarray(data).astype(np.float64)

    if M is None:
        M = median_func(data, axis=axis)
    if axis is not None:
        M = np.expand_dims(M, axis=axis)

    # set up the differences
    d = data - M

    # set up the weighting
    mad = median_absolute_deviation(data, axis=axis, ignore_nan=ignore_nan)

    if axis is None:
        # data is constant or mostly constant OR
        # data contains NaNs and ignore_nan=False
        if mad == 0.0 or np.isnan(mad):
            return mad**2  # variance units
    else:
        mad = np.expand_dims(mad, axis=axis)

    with np.errstate(divide="ignore", invalid="ignore"):
        u = d / (c * mad)

    # now remove the outlier points
    # ignore RuntimeWarnings for comparisons with NaN data values
    with np.errstate(invalid="ignore"):
        mask = np.abs(u) < 1
    if isinstance(mask, np.ma.MaskedArray):
        mask = mask.filled(fill_value=False)  # exclude masked data values

    u = u**2

    if modify_sample_size:
        n = sum_func(mask, axis=axis)
    else:
        # set good values to 1, bad values to 0
        include_mask = np.ones(data.shape)
        if isinstance(data, np.ma.MaskedArray):
            include_mask[data.mask] = 0
        if ignore_nan:
            include_mask[np.isnan(data)] = 0
        n = np.sum(include_mask, axis=axis)

    f1 = d * d * (1.0 - u) ** 4
    f1[~mask] = 0.0
    f1 = sum_func(f1, axis=axis)
    f2 = (1.0 - u) * (1.0 - 5.0 * u)
    f2[~mask] = 0.0
    f2 = np.abs(np.sum(f2, axis=axis)) ** 2

    # If mad == 0 along the specified ``axis`` in the input data, return
    # 0.0 along that axis.
    # Ignore RuntimeWarnings for divide by zero.
    with np.errstate(divide="ignore", invalid="ignore"):
        value = n * f1 / f2
        if np.isscalar(value):
            return value

        where_func = np.where
        if isinstance(data, np.ma.MaskedArray):
            where_func = np.ma.where  # return MaskedArray
        return where_func(mad.squeeze() == 0, 0.0, value)
